/* -*- mode: C; fill-column: 78; indent-tabs-mode: t; tab-width: 8; c-basic-offset: 8; -*- */

#include <stdlib.h>		/* malloc(), free(), exit() */
#include <string.h>		/* memset(), strcmp() */
#include <stdio.h>		/* printf(), fprintf() */
#include <sysexits.h>		/* EX_* */
#include <syslog.h>		/* openlog(), syslog(), closelog() */
#include <sys/types.h>		/* getaddrinfo(), freeaddrinfo(), fork(),
				   open() */
#include <sys/socket.h>		/* getaddrinfo(), freeaddrinfo() */
#include <netdb.h>		/* getaddrinfo(), freeaddrinfo() */
#include <netinet/in.h>		/* struct sockaddr_in */
#include <ctype.h>		/* isspace() */
#include <unistd.h>		/* fork(), setsid(), close() */
#include <sys/stat.h>		/* open() */
#include <fcntl.h>		/* open() */

#include <libmilter/mfapi.h>


#define STDLEN 256	/* long enough string for hostnames to be looked up */

struct rblArray_t
{
	int rblNo;		/* total number of RBLs to check */
	char **suffix;		/* array[rblNo] of suffixes */
	int *suffix_len;	/* array[rblNo] of lengths of suffixes */
};

struct mlfiPriv
{
	/* Array of length rblNo, of boolean, containing the check of the
	   current message for each RBL */
	char *rblStatus;
	char *subject;
	int subject_len;
	char isSpam;
};

const char *myHeaderTitle= "X-Spam-Tests";
struct rblArray_t rblArray;
int myHeaderMaxMem;
/* The following are all options set on the command line */
int debug= 0;
int foreground= 0;
int rejectPositive= 0;
char *subjTag= NULL;
char *sockAddr= NULL;
char *confFile= NULL;



/*
  Returns a pointer to a new memory buffer of length len containing the
  contents of s[0..len-1] and a terminating NULL, or NULL if the memory
  allocation failed.
*/
char *mem2str(const void *s, int len)
{
	char *d;

	d= malloc(len+1);
	if (d != NULL)
	{
		memcpy(d, s, len);
		d[len]= '\0';
	}
	return d;
}

/*
  Allocate necessary memory for the message
*/
sfsistat mlfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
{
	struct mlfiPriv *priv;

	priv= malloc(sizeof(struct mlfiPriv));
	if ( priv==NULL )
		return SMFIS_TEMPFAIL;
	smfi_setpriv(ctx, priv);
	(*priv).rblStatus= malloc(rblArray.rblNo * sizeof(char));
	if ( (*priv).rblStatus==NULL )
		return SMFIS_TEMPFAIL;
	(*priv).subject= NULL;
	
	return SMFIS_CONTINUE;
}

/* 
   Zero the (*priv).rblStatus array in the beginning of the message
*/
sfsistat mlfi_envfrom(SMFICTX *ctx, char **argv)
{
	struct mlfiPriv *priv;

	priv= smfi_getpriv(ctx);
	memset((*priv).rblStatus,'\0', rblArray.rblNo * sizeof(char));

	free((*priv).subject);
	(*priv).subject= NULL;
	(*priv).subject_len= 0;
	(*priv).isSpam= 0;

	return SMFIS_CONTINUE;
}

/*
  For now only store the subject of the message in (*priv).subject
*/
sfsistat mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
{
	struct mlfiPriv *priv;

	priv= smfi_getpriv(ctx);

	/* if we are told to mark the subject (-s switch) */
	if ( subjTag != NULL )
	{
		/* If the _first_ "Subject" header is encountered */
		if ( (strcmp("Subject",headerf) == 0)
		     && ((*priv).subject == NULL )
		     )
		{
			/* then prepend it with subjTag */
			(*priv).subject_len= strlen(headerv) + strlen(subjTag);
			(*priv).subject= malloc( ((*priv).subject_len+1) * sizeof(char) );
			if ( (*priv).subject == NULL )
				return SMFIS_TEMPFAIL;
			strcpy((*priv).subject, subjTag);
			strcat((*priv).subject, headerv);
		}
	}

	return SMFIS_CONTINUE;
}

/*
  Check all RBLs stored in rblArray for the IP address in addr, and modify
  accordingly the header which will be added to the message.
*/
void checkAllRBLs(SMFICTX *ctx, unsigned char addr[4])
{
	struct mlfiPriv *priv;
	char txtAddr[STDLEN];
	struct addrinfo *res= NULL;
	int i, retval;

	priv= smfi_getpriv(ctx);
	/* for every RBL in my list */
	for (i=0; i<rblArray.rblNo; i++)
	{
		retval= snprintf(txtAddr,STDLEN,"%u.%u.%u.%u.%s",
				 addr[3],addr[2],addr[1],addr[0],
				 rblArray.suffix[i]);
		if ( retval<STDLEN )	/* if txtAddr was long enough */
		{
			if (debug==1)
				syslog(LOG_DEBUG, "RBL checking %s...", txtAddr);
			/* Use getaddrinfo(), the reentrant version of
			   gethostbyname() as proposed by POSIX.1-2001 */
			retval= getaddrinfo(txtAddr, NULL, NULL, &res);
			if (retval==0)
			{
				if (debug==1)
					syslog(LOG_DEBUG, "positive!");
				freeaddrinfo(res);
				(*priv).rblStatus[i]= 1;
			}
			else
				if (debug==1)
					syslog(LOG_DEBUG, "negative.");
		}
	}
}

/*
  Repetitive SURBL check of an FQDN until found positive. Then modify
  accordingly the message header.

  Example: check not.spammer.co.uk at multi.surbl.org
  1) not.spammer.co.uk.multi.surbl.org
  2) spammer.co.uk.multi.surbl.org
  3) co.uk.multi.surbl.org

  Since we stress a lot the DNS subsystem, better have a lightweight caching
  nameserver doing the job, or TODO: domain whitelist.
*/
void checkAllSURBLs(SMFICTX *ctx, const char *hostname, int len)
{
	struct mlfiPriv *priv;
	struct addrinfo *res= NULL;
	char txtAddr[STDLEN], *hostnameToCheck;
	int i, j, retval, dotsNo, positive;

	priv= smfi_getpriv(ctx);

	/* Count the dots */
	dotsNo=0;
	for (i=0; i<len; i++)
		if (hostname[i]=='.')
			dotsNo++;

	/* for every RBL in my list */
	for (i=0; i<rblArray.rblNo; i++)
	{
		/* Store in txtAddr the full hostname, prefix+SURBLsuffix */
		retval= snprintf(txtAddr, STDLEN, 
				 "%s.%s", hostname, rblArray.suffix[i]);
		/* if txtAddr was long enough */
		if (retval<STDLEN)
		{
			hostnameToCheck= txtAddr;
			positive= 0;
			j= 0;
			/* for all possible prefixes, excluding the last one
			   (.gr,.com), unless a SURBL check turns out
			   positive */
			while ( (positive==0) && (j<dotsNo) )
			{
				if (debug==1)
					syslog(LOG_DEBUG, "RBL checking %s...",
					       hostnameToCheck);
				retval= getaddrinfo(hostnameToCheck, NULL, NULL, &res);
				if (retval==0)
				{
					if (debug==1)
						syslog(LOG_DEBUG, "positive!");
					freeaddrinfo(res);
					positive= 1;
					(*priv).rblStatus[i]= 1;
				}
				else
				{
					if (debug==1)
						syslog(LOG_DEBUG, "negative.");
					hostnameToCheck= strchr(hostnameToCheck, '.') + 1;
					j++;
				}
			}
		}
	}
}

/*
  Find URIs and SURBLcheck them.
  TODO: How does libmilter handles multiple body chunks? 
  If the separation is random then I should handle the situation
  where a URI is broken in pieces!
  TODO: What if the URI ends at exactly the end of *bodyp?
  TODO: http://www.surbl.org/implementation.html
*/
sfsistat mlfi_body(SMFICTX *ctx, unsigned char *bodyp, size_t len)
{
	struct mlfiPriv *priv;
	static const char URIstart[8]= "http://";
	int i, j, retval, tmpStr_len;
	char *tmpStr;
	unsigned char addr[4];

	priv= smfi_getpriv(ctx);
	
	/* i is a pointer to the current char in body, j is the amount of
	   characters of URIstart that are found */
	i=0; j=0;
	while (i<len)
	{
		/* If we found a complete URIstart matching */
		if ( (j>=7) && (j<STDLEN) ) 
		{
			/* If the URI ends */
			if ( (!isalnum(bodyp[i])) && (strchr(".-_",bodyp[i])==NULL)
			     )
			{
				/* Keep the blah.blue.bleh part */
				tmpStr= mem2str(&bodyp[i-j+7], j-7);
				if ( tmpStr == NULL )
					return SMFIS_TEMPFAIL;
				tmpStr_len= j-7;
				/* Check if tmpStr is an IP address or a
				   hostname */
				/* TODO: gcc issues a warning about "hh"
				   modifier not being supported in ANSI C. So
				   how should I say that the variables
				   following are unsigned chars? */
				retval= sscanf(tmpStr, "%hhu.%hhu.%hhu.%hhu", 
					       &addr[0],&addr[1],&addr[2],&addr[3]);
				/* And accordingly do the checks... */
				if ( retval==4 )
					checkAllRBLs(ctx, addr);
				else
					checkAllSURBLs(ctx, tmpStr, tmpStr_len);
				free(tmpStr);
				/* Restart searching for URIstart matching */
				j= 0;
			}
			else
				j++;
			
		}
		else
		{
			if ( bodyp[i] == URIstart[j] )
				j++;
			else
				j= 0;
		}
		i++;
	}

	return SMFIS_CONTINUE;
}

/*
  Add a myHeader to the message, according to the results stored in
  (*priv).rblStatus array.
*/
sfsistat mlfi_eom(SMFICTX *ctx)
{
	struct mlfiPriv *priv;
	char *msgUID, *myHeader;
	int i, j, rc;

	priv= smfi_getpriv(ctx);
	msgUID= smfi_getsymval(ctx, "{msg_id}");

	/* Add myHeader to the message */
	myHeader= malloc(myHeaderMaxMem);
	if (myHeader==NULL)
		return SMFIS_TEMPFAIL;
	j= 0;
	for (i=0; i<rblArray.rblNo; i++)
	{
		if (i>0)
		{
			myHeader[j]= ',';
			j++;
		}
		memcpy(&myHeader[j], rblArray.suffix[i], rblArray.suffix_len[i]);
		j+= rblArray.suffix_len[i];
		myHeader[j]= ':';
		j++;
		if ( (*priv).rblStatus[i] == 0 )
		{
			memcpy(&myHeader[j], "OK", 2);
			j+= 2;
		}
		else
		{
			(*priv).isSpam= 1;
			memcpy(&myHeader[j], "SPAM", 4);
			j+= 4;
			/* If -j parameter is set, abort all myHeader handling
			   and simply drop the message, with a 550-5.7.1
			   reply */
			if (rejectPositive == 1)
			{
				/* One STDLEN for the rbl suffix and one
				   STDLEN for the rest of the message */
				char replyStr[2*STDLEN];

				free(myHeader);
				strcpy(replyStr, "Email blocked using ");
				strcat(replyStr, rblArray.suffix[i]);
				rc= smfi_setreply(ctx, "550", "5.7.1", replyStr);
				syslog(LOG_DEBUG, "msg_id: %s, %s, Rejecting!", 
				       msgUID, replyStr);
				if (rc == MI_FAILURE)
					return SMFIS_TEMPFAIL;
				return SMFIS_REJECT;
			}
		}
	}
	myHeader[j]= '\0';
	smfi_addheader(ctx, (char*) myHeaderTitle, myHeader);
	/* TODO: check if return value is MI_FAILURE or MI_SUCCESS */
	syslog(LOG_INFO, "msg_id: %s, added %s header: %s", 
	       msgUID, myHeaderTitle, myHeader);
	free(myHeader);

	if ( ((*priv).isSpam == 1) && 
	     (subjTag != NULL) && ((*priv).subject != NULL) )
	{
		smfi_chgheader(ctx, "Subject", 1, (*priv).subject);
		syslog(LOG_DEBUG, "Prepended %s to message subject", subjTag);
	}

	return SMFIS_CONTINUE;
}

/*
  Free whatever I allocated
*/
sfsistat mlfi_close(SMFICTX *ctx)
{
	struct mlfiPriv *priv;

	priv= smfi_getpriv(ctx);
	if (priv!=NULL)
	{
		free((*priv).rblStatus);
		if ((*priv).subject != NULL)
			free((*priv).subject);
		free(priv);
		smfi_setpriv(ctx, NULL);
	}

	if (debug==1)
		syslog(LOG_DEBUG, "Connection closed.");

	return SMFIS_CONTINUE;
}


struct smfiDesc smfilter =
{
	"uri-milter",
	SMFI_VERSION,
	SMFIF_ADDHDRS|SMFIF_CHGHDRS,
	mlfi_connect, 	/* xxfi_connect, */
	NULL, 		/* mlfi_helo, */
	mlfi_envfrom, 	/* mlfi_envfrom, */
	NULL, 		/* mlfi_envrcpt, */
	mlfi_header, 	/* mlfi_header, */
	NULL, 		/* mlfi_eoh, */
	mlfi_body,	/* mlfi_body, */
	mlfi_eom, 	/* mlfi_eom, */
	NULL, 		/* mlfi_abort, */
	mlfi_close, 	/* mlfi_close */
};

void printUsage(char* progName)
{
	printf("Usage:\n");
	printf("%s -p Socket-Addr -c Conf-File [-h] [-s Subject-Tag] [-d] [-f] [-j]\n\n", progName);
	printf("  -p\tListen to the inet/unix socket defined as Socket-Addr\n");
	printf("  -c\tRead one SURBL from each line in Conf-File\n");
	printf("  -h\tHelp, a.k.a. this message\n");
	printf("  -s\tFor each SURBL-positive mail, prepend Subject-Tag to email subject\n");
	printf("  -d\tDebug mode, currently only more verbose output to syslog\n");
	printf("  -f\tForeground mode, do not daemonize\n");
	printf("  -j\tReject each SURBL-positive email with a 550-5.7.1 reply\n");
	printf("\n\n");
}

void parseCmdLine(int argc, char **argv)
{
	int i;
	int helpNeeded= 0;
	int expectingSocket= 0;
	int expectingConfFile= 0;
	int expectingSubjTag= 0;

	for (i=1; i<argc; i++)
	{
		if (expectingSubjTag == 1)
		{
			subjTag= argv[i];
			expectingSubjTag= 0;
			continue;
		}
		if (expectingSocket == 1)
		{
			sockAddr= argv[i];
			expectingSocket= 0;
			continue;
		}
		if (expectingConfFile == 1)
		{
			confFile= argv[i];
			expectingConfFile= 0;
			continue;
		}
		if (strcmp(argv[i],"-d") == 0)
		{
			debug= 1;
			continue;
		}
		if (strcmp(argv[i],"-h") == 0)
		{
			helpNeeded= 1;
			break;
		}
		if (strcmp(argv[i],"-f") == 0)
		{
			foreground= 1;
			continue;
		}
		if (strcmp(argv[i],"-j") == 0)
		{
			rejectPositive= 1;
			continue;
		}
		if (strcmp(argv[i],"-p") == 0)
		{
			expectingSocket= 1;
			continue;
		}
		if (strcmp(argv[i],"-c") == 0)
		{
			expectingConfFile= 1;
			continue;
		}
		if (strcmp(argv[i],"-s") == 0)
		{
			expectingSubjTag= 1;
			continue;
		}
	}
	if ((helpNeeded==1) || (expectingSocket==1) 
	    || (expectingConfFile==1) || (expectingSubjTag==1) 
	    || (sockAddr==NULL) || (confFile==NULL))
	{
		printUsage(argv[0]);
		if (helpNeeded==1)
			exit(EX_OK);
		else
			exit(EX_USAGE);
	}

}

void parseConfFile(char *confFile)
{
	FILE *f;
	char s[STDLEN], *c;
	int i, k;

	f= fopen(confFile,"r");
	if (f == NULL)
	{
		fprintf(stderr,"fopen(confFile) failed\n");
		exit(EX_UNAVAILABLE);
	}
	while ( fgets(s,STDLEN,f) != NULL )
	{
		c= strchr(s,'\n');
		if ( c == NULL )
		{
			fprintf(stderr,
				"Line in conf file longer than %d characters, exiting...",
				STDLEN);
			exit(EX_UNAVAILABLE);
		}
		i= 0;
		/* Pass all gaps */
		while ( isspace(s[i]) && &(s[i])<c )
			i++;
		/* if not comment, then it is a SURBL suffix! */
		if ( (s[i] != '#') && (s[i] != '\n') )
		{
			rblArray.rblNo++;
			rblArray.suffix_len= realloc(rblArray.suffix_len, 
						      rblArray.rblNo * sizeof(int*));
			rblArray.suffix= realloc(rblArray.suffix, 
						  rblArray.rblNo * sizeof(char*));
			if ( rblArray.suffix_len == NULL || rblArray.suffix == NULL )
				exit(EX_UNAVAILABLE);
			/* Copy suffix to rblArray.suffix */
			k= i;
			while ( !isspace(s[i]) && &(s[i])<c )
				i++;
			rblArray.suffix_len[rblArray.rblNo-1]= i-k;
			rblArray.suffix[rblArray.rblNo-1]= mem2str(&s[k],i-k);
			if ( rblArray.suffix[rblArray.rblNo-1] == NULL )
				exit(EX_UNAVAILABLE);
		}
	}
	fclose(f);

	myHeaderMaxMem= 0;
	for (i=0; i<rblArray.rblNo; i++)
	{
		/* plus 4 for the "OK" or "SPAM", 1 for ':' and 1 for ',' or
		   '\0' */
		myHeaderMaxMem+= rblArray.suffix_len[i] + 6;
	}
}

/*
  Standard way to daemonize, fork twice to make sure the final child is not a
  session leader. Then you have to close all open file descriptors to detach
  from the console, also change directory somewhere safe and reset umask.
  TODO: Error checking
*/
int daemonize()
{
        int r;

	r= fork();
        if (r < 0)		/* failed */
		return -1;
	else if (r != 0)	/* parent */
		exit(EX_OK);
	else			/* child */
	{
		setsid();
		r = fork();
		if (r < 0)		/* failed */
			return -1;
		else if (r != 0)	/* parent */
			exit(EX_OK);
		else			/* child */
		{
			close(0); close(1); close(2);	/* close old fds */
			open("/dev/null",O_RDWR);	/* new fd 0 */
			dup(0); dup(0);			/* and fds 1 and 2 */
			umask(0);			/* clear umask */
			chdir("/");			/* change to / */
			return 0;
		}
	}
}


/*
  A milter's main() has to do only 3 things:
  1) Set the listening socket with smfi_setcon()
  2) Register the callbacks table with smfi_register()
  3) Pass control to smfi_main
  
  Simple, isn't it?
*/
int main(int argc, char **argv) 
{
	int rc;

	parseCmdLine(argc, argv);
	parseConfFile(confFile);

	openlog(argv[0], LOG_PID|LOG_NDELAY, LOG_LOCAL6);

	smfi_setconn(sockAddr);

	rc= smfi_register(smfilter);
	if (rc == MI_FAILURE)
	{
		fprintf(stderr,"smfi_register() failed\n");
		exit(EX_UNAVAILABLE);
	}

	if (foreground == 0)
		if (daemonize() != 0)
		{
			fprintf(stderr, "daemonization failed\n");
			exit(EX_UNAVAILABLE);
		}

	rc= smfi_main();
	closelog();
	return rc;
}
